@@ -5,3 +5,7 @@ package util
 func BytesToString(b []byte) string {
 	return string(b)
 }
+
+func StringToBytes(s string) []byte {
+	return []byte(s)
+}
